import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt 
from wizard.io import dump_xyz
import numpy as np 
import random 
import re

class MultiMol(): 
    
    def __init__(self, frames):         
        self.frames = frames

    def select_by_size(self, num):
        select_set = []
        for atoms in self.frames:
            if len(atoms) == num:
                select_set.append(atoms)
        return select_set
    
    def split_by_size(self, num):
        select_set = []
        split_set = []
        for atoms in self.frames:
            if len(atoms) == num:
                select_set.append(atoms)
            else:
                split_set.append(atoms)
        return select_set, split_set

    def select_by_formulars(self, *formulars):
        select_set = []
        symbols_lists = []
        for formula in formulars:
            symbols_lists.append(re.findall('([A-Z][a-z]*)', formula))
        for atoms in self.frames:
            symbols = atoms.get_chemical_symbols()
            for symbols_list in symbols_lists:
                if set(symbols) == set(symbols_list):
                    select_set.append(atoms)
                    break
        return select_set
    
    def select_by_config_type(self, config_type):
        select_set = []
        for atoms in self.frames:
            if atoms.info['config_type'] == config_type:
                select_set.append(atoms)
        return select_set
    
    def select_by_num_of_symbols(self, num):
        select_set = []
        for atoms in self.frames:
            s = atoms.get_chemical_symbols()
            unique_elements = set(s)
            if len(unique_elements) == num:
                select_set.append(atoms)
        return select_set
    
    def split_by_num_of_symbols(self, num):
        select_set = []
        split_set = []
        for atoms in self.frames:
            s = atoms.get_chemical_symbols()
            unique_elements = set(s)
            if len(unique_elements) == num:
                select_set.append(atoms)
            else:
                split_set.append(atoms)
        return select_set, split_set

    def select_by_force_error(self, error_min, error_max = 100):
        select_set = []
        for atoms in self.frames:
            f_1 = np.concatenate(atoms.get_forces())
            f_2 = np.concatenate(atoms.info['forces'])
            diff = abs(f_1 - f_2)
            if np.any((diff > error_min) & (diff < error_max)):
                select_set.append(atoms)
        return select_set
    
    def split_by_force_error(self, error_min, error_max = 100):
        select_set = []
        split_set = []
        for atoms in self.frames:
            f_1 = np.concatenate(atoms.get_forces())
            f_2 = np.concatenate(atoms.info['forces'])
            diff = abs(f_1 - f_2)
            if np.any((diff > error_min) & (diff < error_max)):
                select_set.append(atoms)
            else:
                split_set.append(atoms)
        return select_set, split_set
    
    def select_by_force(self, force_min, force_max):
        select_set = []
        for atoms in self.frames:
            f = np.concatenate(atoms.info['forces'])
            if np.all((f > force_min) & (f < force_max)):
                select_set.append(atoms)
        return select_set
    
    def split_by_force(self, force_min, force_max):
        select_set = []
        split_set = []
        for atoms in self.frames:
            f = np.concatenate(atoms.info['forces'])
            if np.all((f > force_min) & (f < force_max)):
                select_set.append(atoms)
            else:
                split_set.append(atoms)
        return select_set, split_set

    def select_random(self, num):
        select_set = []
        select_set = random.sample(self.frames, num)
        return select_set
    
    def split_random(self, num):
        select_set = []
        split_set = []
        select_set = random.sample(self.frames, num)
        split_set = [i for i in self.frames if i not in select_set]
        return select_set, split_set
    
    def subtract_frames(self, frames):
        select_set = [i for i in self.frames if i not in frames]
        return select_set
    
    def subtract_isolated_atom_energy(self, isolated_atom_energy = {}):
        for atoms in self.frames:
            for atom in atoms:
                atoms.info['energy'] -= isolated_atom_energy[atom.symbol]
    
    def cal_force_by_symbol(self, symbol):
        dft_force = []
        calc_force = []
        for atoms in self.frames:
            for atom in atoms:
                if atom.symbol == symbol:
                    dft_force.append(atoms.info['forces'][atom.index])
                    calc_force.append(atoms.get_forces()[atom.index])
        return np.array([dft_force, calc_force])

    def deform(self, scale = np.arange(0.95, 1.06, 0.05)):
        frames = []
        for atoms in self.frames:
            for s in scale:
                atoms_copy = atoms.copy()
                atoms_copy.set_cell(atoms.get_cell() * s, scale_atoms=True)
                frames.append(atoms_copy)
        return frames
    
    def random_strain(self, ratio = 0.04):
        frames = []
        for atoms in self.frames:
            atoms_copy = atoms.copy()
            strain_matrix = np.eye(3) + 2 * ratio * (np.random.random((3,3)) - 0.5)
            new_cell = np.dot(atoms.get_cell(), strain_matrix)
            atoms_copy.set_cell(new_cell, scale_atoms=True)
            frames.append(atoms_copy)
        return frames
    
    def random_displacement(self, max_displacement = 0.4):
        frames = []
        for atoms in self.frames:
            atoms_copy = atoms.copy()
            atoms_copy.positions += np.random.uniform(-max_displacement, max_displacement, atoms_copy.positions.shape)
            frames.append(atoms_copy)
        return frames
    
    def shuffle_symbols(self):
        for atoms in self.frames:
            s = atoms.get_chemical_symbols()
            random.shuffle(s)
            atoms.set_chemical_symbols(s)

    def dump(self, filename):
        for atoms in self.frames:
            dump_xyz(filename, atoms)

    def plot_force_results(self, calcs, labels = None, e_val = [None, None], f_val = [None, None]):
        plt.rcParams["figure.figsize"] = (12, 6)
        plt.rcParams.update({"font.size": 10, "text.usetex": False})
        fig, axes = plt.subplots(1, 2)
        cmap = plt.get_cmap("tab10")
    
        frames = self.frames
        print(len(frames))  

        label_colors = {}
        if labels is None:
            labels = [str(i) for i in range(len(calcs))]
        for calc, label in zip(calcs, labels):
            e_1, e_2, f_1, f_2 = [], [], [], []
            for atoms in frames:
                atoms.calc = calc
                e_1.append(atoms.get_potential_energy() / len(atoms))
                e_2.append(atoms.info['energy'] / len(atoms))
                f_1.append(atoms.get_forces())
                f_2.append(atoms.info['forces'])
            e_1 = np.array(e_1)
            e_2 = np.array(e_2)
            f_1 = np.concatenate(f_1)
            f_2 = np.concatenate(f_2)
            color = cmap(labels.index(label))
            axes[0].plot(e_2, e_1, ".", markersize=10, label=label, color=color)
            axes[1].plot(f_2, f_1, ".", markersize=10, label=label, color=color)
            if label not in label_colors:
                label_colors[label] = color
            e_rmse = np.sqrt(np.mean((e_1-e_2)**2)) 
            f_rmse = np.sqrt(np.mean((f_1-f_2)**2))
            print(f'{label}_E_rmse: {e_rmse * 1000:.2f} meV/atom')
            print(f'{label}_F_rmse: {f_rmse * 1000:.2f} meV/Å')

        x_min, x_max = axes[0].get_xlim()
        y_min, y_max = axes[0].get_ylim()
        min_val = min(x_min, y_min)
        max_val = max(x_max, y_max)
        axes[0].plot([min_val, max_val], [min_val, max_val], 'k--')

        x_min, x_max = axes[1].get_xlim()
        y_min, y_max = axes[1].get_ylim()
        min_val = min(x_min, y_min)
        max_val = max(x_max, y_max)
        axes[1].plot([min_val, max_val], [min_val, max_val], 'k--')
        
        if e_val[0] is not None and e_val[1] is not None:
            axes[0].set_xlim(e_val)
            axes[0].set_ylim(e_val)
        if f_val[0] is not None and f_val[1] is not None:
            axes[1].set_xlim(f_val)
            axes[1].set_ylim(f_val)
        axes[0].set_xlabel("DFT energy (eV/atom)")
        axes[0].set_ylabel("NEP energy (eV/atom)")
        axes[1].set_xlabel("DFT force (eV/Å)")
        axes[1].set_ylabel("NEP force (eV/Å)")

        handles = [plt.Line2D([0], [0], color=color, marker='o', linestyle='') for label, color in label_colors.items()]
        plt.legend(handles, label_colors.keys())
        plt.savefig("force_results.png")
        plt.show()
        plt.close()


from ase import Atoms
from ase.neighborlist import NeighborList
from ase.optimize import QuasiNewton, FIRE, LBFGS
from ase.constraints import ExpCellFilter, FixedLine
from mpl_toolkits.axes_grid1 import ImageGrid    
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt 
import matplotlib.ticker as ticker 
import numpy as np
import random

def write_run(parameters):
    """
    Write the input parameters to a gpumd file 'run.in'.
    """
    with open('run.in','w') as f:
        for i in parameters:
            f.write(i+'\n')

def dump_xyz(filename, atoms):
    def is_valid_key(key):
        return key in atoms.info and atoms.info[key] is not None and all(v is not None for v in atoms.info[key])
    
    valid_keys = {key: is_valid_key(key) for key in ['stress', 'velocities', 'forces', 'group']}

    with open(filename, 'a') as f:
        Out_string = ""
        Out_string += str(int(len(atoms))) + "\n"
        Out_string += "pbc=\"" + " ".join(["T" if pbc_value else "F" for pbc_value in atoms.get_pbc()]) + "\" "
        Out_string += "Lattice=\"" + " ".join(list(map(str, atoms.get_cell().reshape(-1)))) + "\" "
        if 'energy' in atoms.info and atoms.info['energy'] is not None:
            Out_string += " energy=" + str(atoms.info['energy']) + " "
        if valid_keys['stress']:
            if len(atoms.info['stress']) == 6:
                    virial = -atoms.info['stress'][[0, 5, 4, 5, 1, 3, 4, 3, 2]] * atoms.get_volume()
            else:
                virial = -atoms.info['stress'].reshape(-1) * atoms.get_volume()
            Out_string += "virial=\"" + " ".join(list(map(str, virial))) + "\" "
        Out_string += "Properties=species:S:1:pos:R:3:mass:R:1"
        if valid_keys['velocities']:
            Out_string += ":vel:R:3"
        if valid_keys['forces']:
            Out_string += ":force:R:3"
        if valid_keys['group']:
            Out_string += ":group:I:1"
        if 'config_type' in atoms.info and atoms.info['config_type'] is not None:
            Out_string += " config_type="+ atoms.info['config_type']
        if 'weight' in atoms.info and atoms.info['weight'] is not None:
            Out_string += " weight="+ str(atoms.info['weight'])
        Out_string += "\n"
        for atom in atoms:
            Out_string += '{:2} {:>15.8e} {:>15.8e} {:>15.8e} {:>15.8e}'.format(atom.symbol, *atom.position, atom.mass)
            if valid_keys['velocities']:
                Out_string += ' {:>15.8e} {:>15.8e} {:>15.8e}'.format(*atoms.info['velocities'][atom.index])
            if valid_keys['forces']:
                Out_string += ' {:>15.8e} {:>15.8e} {:>15.8e}'.format(*atoms.info['forces'][atom.index])
            if valid_keys['group']:
                Out_string += ' {}'.format(atoms.info['group'][atom.index])
            Out_string += '\n'
        f.write(Out_string)

def parsed_properties(comment):
    properties_str = comment.split('properties=')[1].split()[0]
    properties = properties_str.split(':')
    parsed_properties = {}
    start = 0
    for i in range(0, len(properties), 3):
        property_name = properties[i]
        property_count = int(properties[i+2])
        parsed_properties[property_name] = slice(start, start + property_count)
        start += property_count
    return parsed_properties

def read_symbols(words_in_line, parsed_properties):
    symbol_slice = parsed_properties['species']
    symbol = words_in_line[symbol_slice]
    symbol = symbol[0].lower().capitalize()
    return symbol

def read_positions(words_in_line, parsed_properties):
    pos_slice = parsed_properties['pos']
    pos = words_in_line[pos_slice]
    return [float(pos[0]), float(pos[1]), float(pos[2])]

def read_mass(words_in_line, parsed_properties):
    if 'mass' in parsed_properties:
        mass_slice = parsed_properties['mass']
        mass = words_in_line[mass_slice]
        return float(mass[0])
    else:
        return None

def read_force(words_in_line, parsed_properties):
    force_key = 'forces' if 'forces' in parsed_properties else 'force'
    if force_key in parsed_properties:
        force_slice = parsed_properties[force_key]
        force = words_in_line[force_slice]
        return [float(force[0]), float(force[1]), float(force[2])]
    else:
        return None
    
def read_group(words_in_line, parsed_properties):
    if 'group' in parsed_properties:
        group_slice = parsed_properties['group']
        group = words_in_line[group_slice]
        return int(group[0])
    else:
        return None

def read_velocity(words_in_line, parsed_properties):
    if 'vel' in parsed_properties:
        vel_slice = parsed_properties['vel']
        vel = words_in_line[vel_slice]
        return [float(vel[0]), float(vel[1]), float(vel[2])]
    else:
        return None

def read_xyz(filename):
    """
    Read the atomic positions and other information from a file in XYZ format.
    """
    frames = []
    with open(filename) as f:
        while True:
            line = f.readline()
            if not line:
                break
            symbols = []
            positions = []
            masses = []
            forces = []
            velocities = []
            group = []
            natoms = int(line.strip())
            comment = f.readline().lower().strip()
            if "pbc=\"" in comment:
                pbc_str = comment.split("pbc=\"")[1].split("\"")[0].strip()
                pbc = [True if pbc_value == "t" else False for pbc_value in pbc_str.split()]
            else:
                pbc = [True, True, True]
            lattice_str = comment.split("lattice=\"")[1].split("\"")[0].strip()
            lattice = [list(map(float, row.split())) for row in lattice_str.split(" ")]
            cell = [lattice[0] + lattice[1] + lattice[2], lattice[3] + lattice[4] + lattice[5], lattice[6] + lattice[7] + lattice[8]]
            if "energy=" in comment:
                energy = float(comment.split("energy=")[1].split()[0])
            else: 
                energy = None
            if "virial=" in comment:
                virials = comment.split("virial=\"")[1].split("\"")[0].strip()
                virials = np.array([float(x) for x in virials.split()]).reshape(3, 3)
                stress = - virials / np.linalg.det(cell)
            else:
                stress = None
            if "config_type=" in comment:
                config_type = comment.split("config_type=")[1].split()[0].strip()
            else:
                config_type = None
            parsed_properties_dict = parsed_properties(comment)
            for _ in range(natoms):
                line = f.readline()
                words_in_line = line.split()
                symbols.append(read_symbols(words_in_line, parsed_properties_dict))
                positions.append(read_positions(words_in_line, parsed_properties_dict))
                masses.append(read_mass(words_in_line, parsed_properties_dict))
                forces.append(read_force(words_in_line, parsed_properties_dict))
                velocities.append(read_velocity(words_in_line, parsed_properties_dict))
                group.append(read_group(words_in_line, parsed_properties_dict))
            frames.append(Atoms(symbols=symbols, positions=positions, masses=masses, cell=cell, pbc=pbc, info={'energy': energy, 'stress': stress, 'forces': forces, 'velocities': velocities, 'group': group, 'config_type': config_type}))
    return frames

def read_restart(filename):
    with open(filename, 'r') as f:
        line = f.readline()
        natoms = int(line.split(' ')[0])
        symbols = []
        positions = []
        masses = []
        velocities = []
        group = []
        comment = f.readline()  
        if "pbc=\"" in comment:
            pbc_str = comment.split("pbc=\"")[1].split("\"")[0].strip()
            pbc = [True if pbc_value == "T" else False for pbc_value in pbc_str.split()]
        else:
            pbc = [True, True, True]
        lattice_str = comment.split("Lattice=\"")[1].split("\"")[0].strip()
        lattice = [list(map(float, row.split())) for row in lattice_str.split(" ")]
        cell = [lattice[0] + lattice[1] + lattice[2], lattice[3] + lattice[4] + lattice[5], lattice[6] + lattice[7] + lattice[8]]
        if "group" in comment:
            for _ in range(natoms):
                line = f.readline()
                symbol, x, y, z, mass, vx, vy, vz, group_info= line.split()[:9]
                symbol = symbol.lower().capitalize()
                symbols.append(symbol)
                positions.append([float(x), float(y), float(z)])
                velocities.append([float(vx), float(vy), float(vz)])
                masses.append(mass)
                group.append(group_info)      
            atoms = Atoms(symbols=symbols, positions=positions, masses=masses, cell=cell, pbc=pbc, info={'velocities': velocities, 'group': group})
        else:
            for _ in range(natoms):
                line = f.readline()
                symbol, x, y, z, mass, vx, vy, vz= line.split()[:8]
                symbol = symbol.lower().capitalize()
                symbols.append(symbol)
                positions.append([float(x), float(y), float(z)])
                velocities.append([float(vx), float(vy), float(vz)])
                masses.append(mass)  
            atoms = Atoms(symbols=symbols, positions=positions, masses=masses, cell=cell, pbc=pbc, info={'velocities': velocities})
    return atoms

def plot_e(ed, en, lim=None, symbol=None):
    fig = plt.figure()
    plt.title("NEP energy vs DFT energy", fontsize=16)
    ed = ed - np.mean(ed)
    en = en - np.mean(en)
    ax = plt.gca()
    ax.set_aspect(1)
    xmajorLocator = ticker.MaxNLocator(5)
    ymajorLocator = ticker.MaxNLocator(5)
    ax.xaxis.set_major_locator(xmajorLocator)
    ax.yaxis.set_major_locator(ymajorLocator)
    
    ymajorFormatter = ticker.FormatStrFormatter('%.1f') 
    xmajorFormatter = ticker.FormatStrFormatter('%.1f') 
    ax.xaxis.set_major_formatter(xmajorFormatter)
    ax.yaxis.set_major_formatter(ymajorFormatter)
    
    ax.set_xlabel('DFT energy (eV/atom)', fontsize=14)
    ax.set_ylabel('NEP energy (eV/atom)', fontsize=14)
    
    ax.spines['bottom'].set_linewidth(3)
    ax.spines['left'].set_linewidth(3)
    ax.spines['right'].set_linewidth(3)
    ax.spines['top'].set_linewidth(3)    
    ax.tick_params(labelsize=16)

    
    plt.plot([np.min(ed), np.max(ed)], [np.min(ed), np.max(ed)],
            color='black',linewidth=3,linestyle='--',)
    plt.scatter(ed, en, zorder=200)
    
    m1 = min(np.min(ed), np.min(en))
    m2 = max(np.max(ed), np.max(en))
    if lim is not None:
        m1 = lim[0]
        m2 = lim[1]
    ax.set_xlim(m1, m2)
    ax.set_ylim(m1, m2)

    ax.set_xlim(m1, m2)
    ax.set_ylim(m1, m2)

    rmse = np.sqrt(np.mean((ed-en)**2))
    plt.text(np.min(ed) * 0.85 + np.max(ed) * 0.15, 
             np.min(en) * 0.15 + np.max(ed) * 0.85,
             "RMSE: {:.3f} eV/atom".format(rmse), fontsize=14)
    if symbol is None:
        plt.savefig('e.png')
    else:
        plt.savefig(f'{symbol}_e.png')
    return fig

def plot_f(fd, fn, lim=None, symbol=None):
    fig = plt.figure()
    ax = plt.gca()
    plt.title("NEP forces vs DFT forces", fontsize=16)
    ax.set_aspect(1)
    xmajorLocator = ticker.MaxNLocator(5)
    ymajorLocator = ticker.MaxNLocator(5)
    ax.xaxis.set_major_locator(xmajorLocator)
    ax.yaxis.set_major_locator(ymajorLocator)
    
    ymajorFormatter = ticker.FormatStrFormatter('%.1f') 
    xmajorFormatter = ticker.FormatStrFormatter('%.1f') 
    ax.xaxis.set_major_formatter(xmajorFormatter)
    ax.yaxis.set_major_formatter(ymajorFormatter)
    
    ax.set_xlabel('DFT forces (eV/A)', fontsize=14)
    ax.set_ylabel('NEP forces (eV/A)', fontsize=14)
    
    ax.spines['bottom'].set_linewidth(2)
    ax.spines['left'].set_linewidth(2)
    ax.spines['right'].set_linewidth(2)
    ax.spines['top'].set_linewidth(2)

    ax.tick_params(labelsize=14)

    ax.set_xlim(np.min(fd), np.max(fd))
    ax.set_ylim(np.min(fn), np.max(fn))

    plt.plot([np.min(fd), np.max(fd)], [np.min(fd), np.max(fd)],
            color='black',linewidth=2,linestyle='--')
    plt.scatter(fd.reshape(-1), fn.reshape(-1), s=2)

    m1 = min(np.min(fd), np.min(fn))
    m2 = max(np.max(fd), np.max(fn))
    if lim is not None:
        m1 = lim[0]
        m2 = lim[1]
    ax.set_xlim(m1, m2)
    ax.set_ylim(m1, m2)

    rmse = np.sqrt(np.mean((fd-fn)**2))
    plt.text(np.min(fd) * 0.85 + np.max(fd) * 0.15, 
             np.min(fn) * 0.15 + np.max(fn) * 0.85,
             "RMSE: {:.3f} eV/A".format(rmse), fontsize=14)
    if symbol is None:
        plt.savefig('f.png')
    else:
        plt.savefig(f'{symbol}_f.png')
    return fig

def plot_v(vd, vn, lim=None, symbol=None):
    fig = plt.figure()
    plt.title("NEP virial vs DFT virial", fontsize=16)
    ax = plt.gca()
    ax.set_aspect(1)
    xmajorLocator = ticker.MaxNLocator(5)
    ymajorLocator = ticker.MaxNLocator(5)
    ax.xaxis.set_major_locator(xmajorLocator)
    ax.yaxis.set_major_locator(ymajorLocator)
    
    ymajorFormatter = ticker.FormatStrFormatter('%.1f') 
    xmajorFormatter = ticker.FormatStrFormatter('%.1f') 
    ax.xaxis.set_major_formatter(xmajorFormatter)
    ax.yaxis.set_major_formatter(ymajorFormatter)
    
    ax.set_xlabel('DFT virial (eV/atom)', fontsize=14)
    ax.set_ylabel('NEP virial (eV/atom)', fontsize=14)
    
    ax.spines['bottom'].set_linewidth(3)
    ax.spines['left'].set_linewidth(3)
    ax.spines['right'].set_linewidth(3)
    ax.spines['top'].set_linewidth(3)    
    ax.tick_params(labelsize=16)

    idx = np.where(vn == -10**6)[0]  
    vd = np.delete(vd, idx) 
    vn = np.delete(vn, idx)

    plt.plot([np.min(vd), np.max(vd)], [np.min(vd), np.max(vd)],
            color='black',linewidth=3,linestyle='--',)
    plt.scatter(vd, vn, zorder=200)
    
    m1 = min(np.min(vd), np.min(vn))
    m2 = max(np.max(vd), np.max(vn))
    if lim is not None:
        m1 = lim[0]
        m2 = lim[1]

    ax.set_xlim(m1, m2)
    ax.set_ylim(m1, m2)
    ax.set_xlim(m1, m2)
    ax.set_ylim(m1, m2)

    rmse = np.sqrt(np.mean((vd-vn)**2))
    plt.text(np.min(vd) * 0.85 + np.max(vd) * 0.15, 
             np.min(vn) * 0.15 + np.max(vd) * 0.85,
             "RMSE: {:.3f} eV/atom".format(rmse), fontsize=14)
    if symbol is None:
        plt.savefig('v.png')
    else:
        plt.savefig(f'{symbol}_v.png')
    return fig


def plot_band_structure(atoms, symbol=None):
    if 'band_dict' not in atoms.info:
        raise ValueError('No band structure data found.')
    band_dict = atoms.info['band_dict']
    labels_path = band_dict['labels_path']
    frequencies = band_dict['frequencies']
    distances = band_dict['distances']
    fig = plt.figure()
    axs = ImageGrid(fig, 111, nrows_ncols=(1, len(labels_path)), axes_pad=0.2, label_mode="L")

    max_freq = max([np.max(fq) for fq in frequencies])
    max_dist = distances[-1][-1]
    xscale = max_freq / max_dist * 1.5
    distances_scaled = [d * xscale for d in distances]

    n = 0
    axs[0].set_ylabel("Frequency", fontsize=14)
    for i, path in enumerate(labels_path):
        axs[i].spines['bottom'].set_linewidth(1.5)
        axs[i].spines['left'].set_linewidth(1.5)
        axs[i].spines['right'].set_linewidth(1.5)
        axs[i].spines['top'].set_linewidth(1.5)
        axs[i].tick_params(labelsize=14)
        xticks = [distances_scaled[n][0]]
        for label in path[:-1]:
            xticks.append(distances_scaled[n][-1])
            axs[i].plot([distances_scaled[n][-1], distances_scaled[n][-1]], 
                        [0, max_freq],
                        linewidth=2,
                        linestyle=":",
                        c='grey')
            axs[i].plot(distances_scaled[n], 
                        frequencies[n], 
                        linewidth=2,
                        c='g')
            n += 1
        axs[i].set_xlim(xticks[0], xticks[-1])
        axs[i].set_xticks(xticks)
        axs[i].set_xticklabels(path)
        axs[i].plot([xticks[0], xticks[-1]], 
                    [0, 0], 
                    linewidth=1,
                    c='black')
        
    if symbol is None:
        plt.savefig('phono.png')
    else:
        plt.savefig(f'{symbol}_phono.png')
    
    plt.close()
    return axs

def Prediction():
    e_1, e_2 = [], []
    v_1, v_2 = [], []
    f_x1, f_y1, f_z1, f_x2, f_y2, f_z2= [], [], [], [], [], []
    
    with open('energy_train.out', 'r') as f:
        lines = f.readlines()
        for line in lines:
            row = list(map(float, line.strip().split()))
            if len(row) == 2:
                e_2.append(row[0])
                e_1.append(row[1])
    
    
    with open('virial_train.out', 'r') as f:
        lines = f.readlines()
        for line in lines:
            row = list(map(float, line.strip().split()))
            if len(row) == 12:
                v_2.append(row[0])
                v_2.append(row[1])
                v_2.append(row[2])
                v_2.append(row[3])
                v_2.append(row[4])
                v_2.append(row[5])
                v_1.append(row[6])
                v_1.append(row[7])
                v_1.append(row[8])
                v_1.append(row[9])
                v_1.append(row[10])
                v_1.append(row[11])
    
    with open('force_train.out', 'r') as f:
        lines = f.readlines()
        for line in lines:
            row = list(map(float, line.strip().split()))
            if len(row) == 6:
                f_x2.append(row[0])
                f_y2.append(row[1])
                f_z2.append(row[2])
                f_x1.append(row[3])
                f_y1.append(row[4])
                f_z1.append(row[5])
    
    e_1 = np.array(e_1)
    e_2 = np.array(e_2)
    e_rmse = np.sqrt(np.mean((e_1-e_2)**2)) 
    
    f_1 = f_x1 + f_y1 + f_z1
    f_2 = f_x2 + f_y2 + f_z2
    f_1 = np.array(f_1)
    f_2 = np.array(f_2)
    f_rmse = np.sqrt(np.mean((f_1-f_2)**2))  
    
    v_1 = np.array(v_1)
    v_2 = np.array(v_2)
    idx = np.where(v_2 == -10**6)[0]  
    v_1 = np.delete(v_1, idx) 
    v_2 = np.delete(v_2, idx)
    v_rmse = np.sqrt(np.mean((v_1-v_2)**2))
    
    plot_e(e_1, e_2)
    plot_f(f_1, f_2)
    plot_v(v_1, v_2)
    
    print(e_rmse)
    print(f_rmse)
    print(v_rmse)

def relax(atoms, fmax = 0.01, steps = 500, model = 'qn', method = 'hydro'):
    if method == 'fixed_line':
        constraint = [FixedLine(atom.index, direction=[0, 0, 1]) for atom in atoms]
        atoms.set_constraint(constraint)
        ucf = atoms
    elif method == 'hydro':
        ucf = ExpCellFilter(atoms, scalar_pressure=0.0, hydrostatic_strain=True) 
    elif method == 'ucf':
        ucf = atoms
    elif method == 'no_opt':
        return
    else:
        raise ValueError('Invalid relaxation method.')
    
    if model == 'qn':
        dyn = QuasiNewton(ucf)
    elif model == 'lbfgs':
        dyn = LBFGS(ucf)
    elif model == 'fire':
        dyn = FIRE(ucf)
    else:
        raise ValueError('Invalid optimization model.')
    
    dyn.run(fmax=fmax, steps=steps)

def get_nth_nearest_neighbor_index(atoms, index, nth):
    cutoffs = [5.0] * len(atoms)
    neighbor_list = NeighborList(cutoffs, self_interaction=False, bothways=True)
    neighbor_list.update(atoms)
    indices, offsets = neighbor_list.get_neighbors(index)
    distances = [atoms.get_distance(index, neighbor) for neighbor in indices]
    sorted_neighbors = sorted(zip(distances, indices), key=lambda x: x[0])
    current_order = 0
    nth_nearest_neighbor_index = None

    for i in range(len(sorted_neighbors)):
        if current_order == nth:
            nth_nearest_neighbor_index = sorted_neighbors[i][1]
            break
        if i < len(sorted_neighbors) - 1 and not np.isclose(sorted_neighbors[i][0], sorted_neighbors[i + 1][0]):
            current_order += 1
    return nth_nearest_neighbor_index

def symbol_to_string(symbols):
    element_counts = {}
    for element in symbols:
        if element in element_counts:
            element_counts[element] += 1
        else:
            element_counts[element] = 1
    result_string = ''
    for element, count in element_counts.items():
        result_string += f'{element}{count}'
    return result_string

def active_learning(frames, main_potential, potentials, error_min, error_max = 100, n = 1000000):
    Train_set = []
    for atoms in frames:
        diff = []
        atoms.calc = main_potential
        f_0 = np.concatenate(atoms.get_forces())
        for potential in potentials:
            atoms.calc = potential
            diff.append(np.concatenate(atoms.get_forces()) - f_0)
        diff_array = np.concatenate(diff)
        if np.any((diff_array > error_min) & (diff_array < error_max)):
            Train_set.append(atoms)
    if len(Train_set) > n:
        Train_set = random.sample(Train_set, n)
    return Train_set

def plot_thermo_out(filenames, column=2, num=1):
    mpl.rcParams['font.size'] = 16
    plt.figure(figsize=(8, 6))
    fig, ax = plt.subplots()
    for filename in filenames:
        with open(filename, 'r') as f:
            lines = f.readlines()
        lines = [line.strip() for line in lines]
        lines = [line.split() for line in lines]
        lines = [line for line in lines if len(line) == 12]  
        lines = [[float(i) for i in line] for line in lines]
        data = np.array(lines)
        thermo = data[:, column] / num
        ax.plot(thermo, label=filename)
    ax.yaxis.set_major_formatter('{:.3f}'.format) 
    ax.set_xlabel('step(100)')  
    ax.set_ylabel('eV/atom')
    ax.legend()
    plt.tight_layout() 
    plt.savefig('thermo.png')
    return thermo

def plot_training_result(dirname = '', type = 'train'):
    plt.rcParams["figure.figsize"] = (6, 6)
    plt.rcParams.update({"font.size": 10, "text.usetex": False})
    fig, axes = plt.subplots(2, 2)

    loss = np.loadtxt(dirname + "loss.out")
    energy = np.loadtxt(dirname + "energy_" + type + ".out")
    force = np.loadtxt(dirname + "force_" + type + ".out")
    virial_initial = np.loadtxt(dirname + "virial_" + type + ".out")
    virial= virial_initial[np.logical_not(np.any(virial_initial == -1e6, axis=1))]
    loss = loss[:, 2:7]

    axes[0, 0].loglog(loss)
    axes[0, 0].annotate("(a)", xy=(0.0, 1.1), xycoords="axes fraction", va="top", ha="right")
    axes[0, 0].set_xlabel("Generation")
    axes[0, 0].set_ylabel("Loss functions")
    axes[0, 0].legend(["L1", "L2", "Energy", "Force", "Virial"], loc="lower left")

    axes[0, 1].plot(energy[:, 1], energy[:, 0], ".", markersize=10, color=[0, 0.45, 0.74])
    axes[0, 1].plot(axes[0, 1].get_xlim(), axes[0, 1].get_xlim(), linewidth=2, color=[0.85, 0.33, 0.1])
    axes[0, 1].annotate("(b)", xy=(0.0, 1.1), xycoords="axes fraction", va="top", ha="right")
    axes[0, 1].set_xlabel("DFT energy (eV/atom)")
    axes[0, 1].set_ylabel("NEP energy (eV/atom)")
    x_min = np.min(energy[:, 1]) * 1.1
    x_max = np.max(energy[:, 1]) * 1.1
    y_min = np.min(energy[:, 1]) * 1.1
    y_max = np.max(energy[:, 1]) * 1.1

    axes[0, 1].set_xlim(x_min, x_max)
    axes[0, 1].set_ylim(y_min, y_max)

    axes[1, 0].plot(force[:, 3:6], force[:, 0:3], ".", markersize=10, color=[0, 0.45, 0.74])
    axes[1, 0].plot(axes[1, 0].get_xlim(), axes[1, 0].get_xlim(), linewidth=2, color=[0.85, 0.33, 0.1])
    axes[1, 0].annotate("(c)", xy=(0.0, 1.1), xycoords="axes fraction", va="top", ha="right")
    axes[1, 0].set_xlabel("DFT force (eV/Å)")
    axes[1, 0].set_ylabel("NEP force (eV/Å)")
    x_min = np.min(force[:, 3:6]) * 1.1
    x_max = np.max(force[:, 3:6]) * 1.1
    y_min = np.min(force[:, 3:6]) * 1.1
    y_max = np.max(force[:, 3:6]) * 1.1

    axes[1, 0].set_xlim(x_min, x_max)
    axes[1, 0].set_ylim(y_min, y_max)

    cols = virial.shape[1]
    half = cols // 2
    axes[1, 1].plot(virial[:, :half], virial[:, half:], ".", markersize=10, color=[0, 0.45, 0.74])
    axes[1, 1].plot(axes[1, 1].get_xlim(), axes[1, 1].get_xlim(), linewidth=2, color=[0.85, 0.33, 0.1])
    axes[1, 1].annotate("(d)", xy=(0.0, 1.1), xycoords="axes fraction", va="top", ha="right")
    axes[1, 1].set_xlabel("DFT virial (eV/atom)")
    axes[1, 1].set_ylabel("NEP virial (eV/atom)")
    x_min = np.min(virial[:, 0:6]) * 1.1
    x_max = np.max(virial[:, 0:6]) * 1.1
    y_min = np.min(virial[:, 0:6]) * 1.1
    y_max = np.max(virial[:, 0:6]) * 1.1

    axes[1, 1].set_xlim(x_min, x_max)
    axes[1, 1].set_ylim(y_min, y_max)

    plt.subplots_adjust(hspace=0.4, wspace=0.3)

    plt.savefig("train_results.png")
    plt.show()

def plot_force_results(frames, calcs, labels = None, e_val = [None, None], f_val = [None, None]):
    plt.rcParams["figure.figsize"] = (12, 5)
    plt.rcParams.update({"font.size": 18, "text.usetex": False})
    fig, axes = plt.subplots(1, 2) 
    plt.subplots_adjust(wspace=0.3, bottom=0.2)
    cmap = plt.get_cmap("tab10")
  
    print(len(frames))  

    label_colors = {}
    if labels is None:
        labels = [str(i) for i in range(len(calcs))]
    for calc, label in zip(calcs, labels):
        e_1, e_2, f_1, f_2 = [], [], [], []
        for atoms in frames:
            atoms.calc = calc
            e_1.append(atoms.get_potential_energy() / len(atoms))
            e_2.append(atoms.info['energy'] / len(atoms))
            f_1.append(atoms.get_forces())
            f_2.append(atoms.info['forces'])
        e_1 = np.array(e_1)
        e_2 = np.array(e_2)
        f_1 = np.concatenate(f_1)
        f_2 = np.concatenate(f_2)
        color = cmap(labels.index(label))
        axes[0].plot(e_2, e_1, ".", markersize=10, label=label, color=color)
        axes[1].plot(f_2, f_1, ".", markersize=10, label=label, color=color)
        if label not in label_colors:
            label_colors[label] = color
        e_rmse = np.sqrt(np.mean((e_1-e_2)**2)) 
        f_rmse = np.sqrt(np.mean((f_1-f_2)**2))
        print(f'{label}_E_rmse: {e_rmse * 1000:.2f} meV/atom')
        print(f'{label}_F_rmse: {f_rmse * 1000:.2f} meV/Å')

    x_min, x_max = axes[0].get_xlim()
    y_min, y_max = axes[0].get_ylim()
    min_val = min(x_min, y_min)
    max_val = max(x_max, y_max)
    axes[0].plot([min_val, max_val], [min_val, max_val], 'k--')

    x_min, x_max = axes[1].get_xlim()
    y_min, y_max = axes[1].get_ylim()
    min_val = min(x_min, y_min)
    max_val = max(x_max, y_max)
    axes[1].plot([min_val, max_val], [min_val, max_val], 'k--')
    
    if e_val[0] is not None and e_val[1] is not None:
        axes[0].set_xlim(e_val)
        axes[0].set_ylim(e_val)
    if f_val[0] is not None and f_val[1] is not None:
        axes[1].set_xlim(f_val)
        axes[1].set_ylim(f_val)
    axes[0].set_xlabel("DFT energy (eV/atom)")
    axes[0].set_ylabel("NEP energy (eV/atom)")
    axes[1].set_xlabel("DFT force (eV/Å)")
    axes[1].set_ylabel("NEP force (eV/Å)")
    axes[0].text(0.05, 0.95, '(a)', transform=axes[0].transAxes, verticalalignment='top')
    axes[1].text(0.05, 0.95, '(b)', transform=axes[1].transAxes, verticalalignment='top')

    handles = [plt.Line2D([0], [0], color=color, marker='o', linestyle='') for label, color in label_colors.items()]
    plt.legend(handles, label_colors.keys())
    plt.savefig("force_results.png")
    plt.show()
    plt.close()


from ase import units
from ase.lattice.cubic import FaceCenteredCubic
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md.verlet import VelocityVerlet
from ase.md.langevin import Langevin
from ase.md.nptberendsen import NPTBerendsen
from ase.io.trajectory import Trajectory

class MolecularDynamics:
    def __init__(self, atoms, calc, timestep = units.fs, dump_interval = 100, report_interval = 100):
        self.atoms = atoms
        self.calc = calc
        self.timestep = timestep
        self.dump_interval = dump_interval
        self.report_interval = report_interval
        self.nsteps = 0

    def report(self):  
        atoms = self.atoms
        self.nsteps += self.report_interval
        epot = atoms.get_potential_energy() / len(atoms)
        ekin = atoms.get_kinetic_energy() / len(atoms)
        print('Step: %d Energy per atom: Epot = %.3feV  Ekin = %.3feV (T=%3.0fK)  '
            'Etot = %.3feV' % (self.nsteps, epot, ekin, ekin / (1.5 * units.kB), epot + ekin))

    def NVE(self, temperature_K = 300, steps = 10000):
        atoms = self.atoms
        atoms.calc = self.calc
        traj = Trajectory('NVE.traj', 'w', atoms)
        if 'velocities' not in self.atoms.info or self.atoms.info['velocities'] is None:
            MaxwellBoltzmannDistribution(self.atoms, temperature_K=temperature_K)
        dyn = VelocityVerlet(self.atoms, self.timestep)
        dyn.attach(self.report, interval=self.report_interval)
        dyn.attach(traj.write, interval=self.dump_interval)
        dyn.run(steps)

    def NVT(self, temperature_K = 300, steps = 10000):
        atoms = self.atoms
        atoms.calc = self.calc
        traj = Trajectory('NVT.traj', 'w', atoms)
        if 'velocities' not in self.atoms.info or self.atoms.info['velocities'] is None:
            MaxwellBoltzmannDistribution(self.atoms, temperature_K=temperature_K)
        dyn = Langevin(self.atoms, self.timestep, temperature_K, 0.02)
        dyn.attach(self.report, interval=self.report_interval)
        dyn.attach(traj.write, interval=self.dump_interval)
        dyn.run(steps)

    def NPT(self, temperature_K = 300, pressure = 0, steps = 10000):
        atoms = self.atoms
        atoms.calc = self.calc
        traj = Trajectory('NPT.traj', 'w', atoms)
        if 'velocities' not in self.atoms.info or self.atoms.info['velocities'] is None:
            MaxwellBoltzmannDistribution(self.atoms, temperature_K=temperature_K)
        dyn = NPTBerendsen(self.atoms, timestep=self.timestep, temperature_K=temperature_K,
                taut=100 * units.fs, pressure=pressure,
                taup=1000 * units.fs, compressibility_au=4.57e-5 / units.bar)
        dyn.attach(self.report, interval=self.report_interval)
        dyn.attach(traj.write, interval=self.dump_interval)
        dyn.run(steps)




import traceback
import numpy as np
import spglib
import ase
from ase import Atoms
from ase.cell import Cell
from phonopy import Phonopy
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.phonon.band_structure import get_band_qpoints_and_path_connections
from phonopy.interface.phonopy_yaml import PhonopyYaml

def ase2phono(atoms):
    return PhonopyAtoms(symbols=atoms.get_chemical_symbols(),
                        cell=atoms.cell.array,
                        scaled_positions=atoms.get_scaled_positions())

def phono2ase(cell):
    return Atoms(symbols=cell.get_chemical_symbols(),
                    scaled_positions=cell.get_scaled_positions(),
                    cell=cell.get_cell(),
                    pbc=True)

class PhonoCalc:
    def __init__(self, atoms, calc, dim='Auto', mesh=(10, 10, 10), t_step=10, t_max=0., t_min=0.):
        self.atoms = atoms
        self.calc = calc
        self.dim = dim
        self.mesh = mesh
        self.t_step = t_step
        self.t_max = t_max
        self.t_min = t_min
        self.results = {}
        try:
            print('Calculating force constants...')
            unitcell = ase2phono(self.atoms)
            if isinstance(self.dim, str):
                if self.dim == 'Auto':
                    lengths = self.atoms.cell.lengths()
                    supercell_matrix = np.round(10 / lengths).astype('int')
            else:
                supercell_matrix = self.dim

            self.phonon = Phonopy(
                unitcell=unitcell, 
                supercell_matrix=supercell_matrix,
                primitive_matrix='auto',
                )
            
            self.phonon.generate_displacements(distance=0.01)
            supercells = self.phonon.get_supercells_with_displacements()
            set_of_forces = []
            for cell in supercells:
                forces = self.calc.get_forces(phono2ase(cell))
                forces -= np.mean(forces, axis=0)
                set_of_forces.append(forces)
            set_of_forces = np.array(set_of_forces)
            self.phonon.produce_force_constants(forces=set_of_forces)
            self.results['force_constants'] = self.phonon.force_constants

            phpy_yaml = PhonopyYaml(settings=
                    {'force_sets': True,
                     'displacements': True,
                     'force_constants': True,
                     'born_effective_charge': True,
                     'dielectric_constant': True})
            phpy_yaml.set_phonon_info(self.phonon)

        except Exception as e:
            print(traceback.format_exc())
            print('Fail to collect force constants')

    def get_band_structure(self):
        lattice = self.atoms.get_cell().T
        positions = self.atoms.get_scaled_positions()
        numbers = self.atoms.get_atomic_numbers()
        atoms_cell = (lattice, positions, numbers)
        cell = Cell(spglib.find_primitive(atoms_cell)[0])
        special_points = cell.get_bravais_lattice().get_special_points()
        labels_path = ase.dft.kpoints.parse_path_string(cell.bandpath().path)
        labels, path = [], []
        for label_path in labels_path:
            p = []
            for l in label_path:
                labels.append(l)
                p.append(special_points[l].tolist())
            path.append(p)
        qpoints, connections = get_band_qpoints_and_path_connections(path, npoints=51)
        self.phonon.run_band_structure(qpoints, path_connections=connections, labels=labels)
        bands_dict = self.phonon.get_band_structure_dict()
        bands_dict['labels_path'] = labels_path
        self.results['band_dict'] = bands_dict
        self.atoms.info.update(self.results)
 


from ase import Atoms, Atom
from ase.build import bulk
from ase.optimize import QuasiNewton, FIRE, LBFGS
from ase.constraints import ExpCellFilter, FixedLine
from wizard.io import get_nth_nearest_neighbor_index, dump_xyz, write_run
import numpy as np
import random
import os
import re
import shutil

class SymbolInfo:
    def __init__(self, formula, structure, *lattice_constant):
        self.formula = formula
        self.structure = structure
        self.lattice_constant = lattice_constant
        self.symbols = []
        self.compositions = []
        for symbol, composition in re.findall('([A-Z][a-z]*)(\d*)', formula):
            self.symbols.append(symbol)
            self.compositions.append(int(composition) if composition else 1)
    
    def create_bulk_atoms(self, supercell = (3, 3, 3)):
        symbol, structure, lc = self.symbols[0], self.structure, self.lattice_constant
        if structure == 'hcp':
            atoms = bulk(symbol, structure, a = lc[0], c = lc[1]) * supercell
        else:
            atoms = bulk(symbol, structure, a = lc[0], cubic=True) * supercell
        if len(self.symbols) > 1:
            if len(atoms) < sum(self.compositions):
                raise ValueError('The number of atoms in the unit cell is less than the number of symbols.')
            element_ratio = np.array(self.compositions) / sum(self.compositions)
            element_counts = np.ceil(element_ratio * len(atoms)).astype(int)
            symbols = np.repeat(self.symbols, element_counts)
            np.random.shuffle(symbols)
            atoms.set_chemical_symbols(symbols[:len(atoms)])
        return atoms
    
class Morph():
    def __init__(self, atoms):
        if not isinstance(atoms, Atoms):
            raise TypeError("atoms must be an instance of ase.Atoms")
        self.atoms = atoms
        
    def relax(self, fmax = 0.01, steps = 500, model = 'qn', method = 'hydro'):
        atoms = self.atoms
        if method == 'fixed_line':
            constraint = [FixedLine(atom.index, direction=[0, 0, 1]) for atom in atoms]
            atoms.set_constraint(constraint)
            ucf = atoms
        elif method == 'hydro':
            ucf = ExpCellFilter(atoms, scalar_pressure=0.0, hydrostatic_strain=True) 
        elif method == 'ucf':
            ucf = atoms
        else:
            raise ValueError('Invalid relaxation method.')
        
        if model == 'qn':
            dyn = QuasiNewton(ucf)
        elif model == 'lbfgs':
            dyn = LBFGS(ucf)
        elif model == 'fire':
            dyn = FIRE(ucf)
        elif model == 'no_opt':
            return
        else:
            raise ValueError('Invalid optimization model.')
        
        dyn.run(fmax=fmax, steps=steps)

    def gpumd(self, dirname = 'relax', run_in = ['potential nep.txt', 'velocity 300', 'time_step 1', 
             'ensemble npt_scr 300 300 200 0 500 2000', 'dump_thermo 1000', 'dump_restart 30000', 
             'dump_exyz 10000','run 30000'], nep_path = 'nep.txt', write_in = False):
        atoms = self.atoms
        if os.path.exists(dirname):
            raise FileExistsError('Directory already exists')
        os.makedirs(dirname)
        if os.path.exists(nep_path):
            shutil.copy(nep_path, dirname)
        else:
            raise FileNotFoundError('nep.txt does not exist')
        original_directory = os.getcwd()
        os.chdir(dirname)
        write_run(run_in)
        dump_xyz('model.xyz', atoms)
        if write_in:
           pass
        else:
            os.system('gpumd')
        os.chdir(original_directory)

    def set_pka(self, energy, direction, index = None, symbol = None):
        atoms = self.atoms
        if atoms.info['velocities'] is None:
            raise ValueError('The velocities of atoms are not set.')
        
        if index is None:
            center = np.diag(atoms.get_cell()) / 2.0
            if symbol is None:
                index = np.argmin(np.sum((atoms.positions - center)**2, axis=1))
            else:
                element_indices = [i for i, atom in enumerate(atoms) if atom.symbol == symbol]
                element_positions = atoms.positions[element_indices]
                index = element_indices[np.argmin(np.sum((element_positions - center)**2, axis=1))]

        mass = atoms[index].mass
        vx = pow(2 * energy / mass , 0.5) * direction[0] / pow(np.sum(direction ** 2), 0.5) / 10.18
        vy = pow(2 * energy / mass , 0.5) * direction[1] / pow(np.sum(direction ** 2), 0.5) / 10.18
        vz = pow(2 * energy / mass , 0.5) * direction[2] / pow(np.sum(direction ** 2), 0.5) / 10.18
        delta_momentum = (np.array(atoms.info['velocities'][index]) - np.array([vx, vy, vz])) * mass / (len(atoms) - 1)
        
        atoms_masses = np.array(atoms.get_masses())
        atoms.info['velocities'] += delta_momentum / atoms_masses[:, np.newaxis]
        atoms.info['velocities'][index] = [vx, vy, vz]
        
    def velocity(self, vx, vy, vz, group = 0):
        atoms = self.atoms
        if atoms.info['velocities'] is None:
            raise ValueError('The velocities of atoms are not set.')
        for index in range(len(atoms)):
            if int(atoms.info['group'][index]) == group:
                atoms.info['velocities'][index] = [vx, vy, vz]
        
        atoms_masses = np.array(atoms.get_masses())
        momentum = np.sum(atoms.info['velocities'] * atoms_masses[:, np.newaxis], axis=0) / len(atoms)
        atoms.info['velocities'] -= momentum / atoms_masses[:, np.newaxis]
    
    def shuffle_symbols(self):
        atoms = self.atoms
        s = atoms.get_chemical_symbols()
        random.shuffle(s)
        atoms.set_chemical_symbols(s)

    def prop_element_set(self, symbols = []):
        atoms = self.atoms
        n_atoms = len(atoms)
        n_sym = len(symbols)
        n_repeat, n_extra = divmod(n_atoms, n_sym)
        sym = np.array(symbols * n_repeat + symbols[:n_extra], dtype=str)
        np.random.shuffle(sym)
        atoms.set_chemical_symbols(sym.tolist())

    def element_random_replace(self, symbol1, symbol2, num):
        atoms = self.atoms
        symbols = atoms.get_chemical_symbols()
        indices = np.where(np.array(symbols) == symbol1)[0]
        if len(indices) < num:
            raise ValueError('The number of atoms to be replaced is greater than the number of atoms of the first element.')
        np.random.shuffle(indices)
        indices = indices[:num]
        for index in indices:
            atoms[index].symbol = symbol2

    def coord_element_set(self, coord, symbol):
        atoms = self.atoms
        for atom in atoms:
            if np.allclose(atom.position, coord):
                atom.symbol = symbol
                break

    def scale_lattice(self, scale):
        atoms = self.atoms
        origin_cell = atoms.cell.copy()
        atoms.set_cell(scale * origin_cell, scale_atoms=True)

    def coord_vac_set(self, coord):
        atoms = self.atoms.copy()
        for atom in atoms:
            if np.allclose(atom.position, coord):
                index = atom.index
                break
        del atoms[index]

    def create_interstitial(self, new_atom):
        atoms = self.atoms
        atoms.append(new_atom)

    def create_self_interstitial_atom(self, vector, symbol = None, index = 0):
        atoms = self.atoms
        if symbol is not None:
            atom = Atom(symbol, atoms[index].position - vector)
        else:
            atom = Atom(atoms[index].symbol, atoms[index].position - vector)
        atoms[index].position += vector
        atoms.append(atom)

    def create_di_self_interstitial_atoms(self, vector1, vector2, symbol1 = None, symbol2 = None, index = 0, nth = 1):
        atoms = self.atoms
        index_neibor = get_nth_nearest_neighbor_index(self.atoms, index, nth)
        if symbol1 is not None:
            atom1 = Atom(symbol1, atoms[index].position - vector1)
        else:
            atom1 = Atom(atoms[index].symbol, atoms[index].position - vector1)
        if symbol2 is not None:
            atom2 = Atom(symbol2, atoms[index_neibor].position - vector2)
        else:
            atom2 = Atom(atoms[index_neibor].symbol, atoms[index_neibor].position - vector2)
        atoms[index].position += vector1
        atoms[index_neibor].position += vector2
        atoms.append(atom1)
        atoms.append(atom2)

    def create_vacancy(self, index = 0):
        del self.atoms[index]

    def create_divacancies(self, index1 = 0, nth = 1):
        index2 = get_nth_nearest_neighbor_index(self.atoms, index1, nth)
        if index2 < index1:
            index1, index2 = index2, index1
        del self.atoms[index2]
        del self.atoms[index1]

    def create_vacancies(self, num_vacancies):
        atoms = self.atoms
        if num_vacancies > len(atoms):
            raise ValueError("num_vacancies should be less than or equal to the total number of atoms.")
        indices_to_remove = np.random.choice(len(atoms), num_vacancies, replace=False)
        removed_atoms = atoms[indices_to_remove]
        indices_to_remove = sorted(indices_to_remove, reverse=True)
        for index in indices_to_remove:
            del self.atoms[index]
        return removed_atoms
    
    def insert_atoms(self, atoms_to_insert, distance=1.2):
        indices_to_insert = np.random.choice(len(self.atoms), len(atoms_to_insert), replace=False)
        target_atoms = self.atoms[indices_to_insert]
        for atom_to_insert, target_atom in zip(atoms_to_insert, target_atoms):
            unit_vector = np.random.randn(3)  # Generate a random vector
            unit_vector /= np.linalg.norm(unit_vector)  # Normalize the vector to get a unit vector
            displacement_vector = distance * unit_vector  # Scale the unit vector to the desired distance
            new_position = target_atom.position + displacement_vector  # Add the displacement vector to the target atom's position
            atom_to_insert.position = new_position
            self.atoms.append(atom_to_insert)
            
    def create_fks(self, num_vacancies):
        removed_atoms = self.create_vacancies(num_vacancies)
        self.insert_atoms(removed_atoms)
    
    def get_atoms(self):
        return self.atoms.copy()
    
    def get_potential_energy(self):
        return self.atoms.get_potential_energy()


import setuptools

with open("README.md", "r",encoding="utf-8") as f:
    long_description = f.read()
    
setuptools.setup(
    name = "gpumd-wizard",
    version = "0.0.1",
    author = "Jiahui Liu",
    author_email="jiahui.liu.willow@gmail.com",
    description="Material structure processing software ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Jonsnow-willow/GPUMD-Wizard",                                         
    packages=setuptools.find_packages(),     
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=['ase',
                      'phonopy',
                      'calorine'],
    )

